//
// example 226MB file https://eicaidata.s3.amazonaws.com/2022-09-04_pgun_e-_podio-0.15_edm4hep-0.6_0-30GeV_alldir_1k.edm4hep.root
//
// The following will merge one event from each file into a single output file name "podio_merged.root".
// This example uses the same file twice, but in a real scenario, they would be different files.
//
// podio_merge -i 2022-09-04_pgun_e-_podio-0.15_edm4hep-0.6_0-30GeV_alldir_1k.edm4hep.root -n 1 -i 2022-09-04_pgun_e-_podio-0.15_edm4hep-0.6_0-30GeV_alldir_2k.edm4hep.root
//

#include <iostream>
#include <iomanip>
#include <ctime>
#include <set>
#include <filesystem>

#include <podio/ROOTReader.h>
#include <podio/ROOTWriter.h>
#include <podio/EventStore.h>
#include <podio/CollectionBase.h>

#include <TFile.h>

#include "datamodel_glue.h"  // This file generated by make_datamodel_glue.py (see CMakeLists.txt)

//------------------------------------------------------------------------------
// CopyCollectionT
//
// Copy the named collection from the input store to the output store.
//------------------------------------------------------------------------------
template<typename T, typename C>
podio::CollectionBase* CopyCollectionT(podio::EventStore& store_in, podio::EventStore& store_out, const std::string& name) {

    // Get pointer to collection in store_in.
    const C* collection_in=nullptr;
    store_in.get<C>(name, collection_in);

    // Get pointer to collection in store_out. Create if neccessary.
    // (this seems like I'm doing something naughty.)
    C* collection_out=nullptr;
    const C* tmpPtr = nullptr;
    if( store_out.get<C>(name, tmpPtr) ) {
        collection_out = const_cast<C*>(tmpPtr);
    }else{
        auto &c = store_out.create<C>( name );
        collection_out = &c;
    }

    // Loop over objects in input collection and create independent copies
    std::vector<T*> objects;
    for( const auto &t : *collection_in ) {
        if (t.isAvailable()) {
            collection_out->push_back( t.clone() );
        }
    }

    return collection_out;
}

//------------------------------------------------------------------------------
// Usage
//------------------------------------------------------------------------------
void Usage(std::string mess=""){

    std::cout << std::endl << "Usage: podio_merge [-m max_events] -i file_signal -n 10 -i file_background -o file_out" << std::endl << std::endl;
    if( !mess.empty() ) std::cout << mess << std::endl << std::endl;
}

//------------------------------------------------------------------------------
// Merge
//
// Copy all collections from store_in tp store_out.
//------------------------------------------------------------------------------
void Merge( podio::EventStore *store_in, podio::EventStore *store_out){
    auto collectionIdTable = store_in->getCollectionIDTable();
    for( auto id : collectionIdTable->ids()){
        auto name = collectionIdTable->name(id);
        podio::CollectionBase *coll = nullptr;
        if (store_in->get(id, coll)) {
            auto type = coll->getValueTypeName();

            // The following will end up calling CopyCollectionT above by way of datamodel_glue.h
            CopyCollection(*store_in, *store_out, name, type);
        }
    }
}

//------------------------------------------------------------------------------
// main
//------------------------------------------------------------------------------
int main(int narg, char *argv[]) {

    // Extract files to merge from command line
    std::vector<std::tuple<std::string, int64_t>> in_files;
    std::string out_file = "podio_merged.root";
    int64_t num_to_merge = 1;
    int64_t max_events = 1000000000;
    for (int iarg = 1; iarg < narg; iarg++) {
        std::string arg = argv[iarg];
        std::string next = (iarg + 1) < narg ? argv[iarg + 1] : "";
        if (arg == "-i" && !next.empty()) in_files.emplace_back(next, num_to_merge);
        if (arg == "-n" && !next.empty()) num_to_merge = atoi(next.c_str());
        if (arg == "-o" && !next.empty()) out_file = next;
        if (arg == "-m" && !next.empty()) max_events = atoi(next.c_str());
    }
    if (in_files.size() < 2) {
        Usage("Must supply at least 2 input files");
        exit(0);
    }

    // Open input files for reading
    std::vector<std::tuple<podio::ROOTReader *, podio::EventStore *, int64_t, int64_t> > readers; // values are: Nevents to merge, events read from file
    for (auto &[filename, Nevents]: in_files) {
        if (!std::filesystem::exists(filename)) { throw std::runtime_error("Input file does not exist!"); }
        auto reader = new podio::ROOTReader;
        reader->openFile(filename);
        if (!reader->isValid())
            throw std::runtime_error("Unable to open podio file for reading. podio ROOTReader says it is invalid");
        auto store = new podio::EventStore;
        store->setReader(reader);
        readers.emplace_back(reader, store, Nevents, 0);
    }
    std::cout << "Opened " << readers.size() << " input files." << std::endl;

    // Open output file for writing
    podio::EventStore store_out; // EventStore used to write events out
    podio::ROOTWriter rootwriter(out_file, &store_out);

    // Loop over all events in first file
    if (std::get<0>(readers[0])->getEntries() < max_events) max_events = std::get<0>(readers[0])->getEntries();
    std::cout << "Merging into " << max_events << " events" << std::endl;
    auto start_time = clock();
    int64_t ievent = 0;
    for (; ievent < max_events; ievent++) {
        auto event_start_time = clock();

        // Clear the output store
        store_out.clear();

        // Loop over input files
        for (auto &[reader, store, num_input_events, nevents_read]: readers) {
            for (int imerge = 0; imerge < num_input_events; imerge++) {
                if (++nevents_read >= reader->getEntries())nevents_read = 0; // reset to start of file
                store->clear();
                reader->endOfEvent();
                reader->goToEvent(nevents_read);
                Merge(store, &store_out);
            }
        }

        // Make sure all collections are written out
        if (ievent == 0) {
            auto collectionIdTable = store_out.getCollectionIDTable();
            for (auto id: collectionIdTable->ids()) {
                auto name = collectionIdTable->name(id);
                rootwriter.registerForWrite(name);
            }
        }

        // Write the event
        rootwriter.writeEvent();

        // Print status
        auto rate = CLOCKS_PER_SEC / (double) (clock() - event_start_time);
        auto avg_rate = CLOCKS_PER_SEC / (double) (clock() - start_time) * (double) (ievent + 1);
        std::cout << " " << ievent + 1 << " events copied  ( " << std::setprecision(2) << rate << " Hz  avg. "
                  << avg_rate << " Hz)      \r";
        std::cout.flush();
    }
    std::cout << std::endl << "Merged into " << ievent << " events" << std::endl;

    // Close output file
    rootwriter.finish();

    // Close input files
    for (auto &[reader, store, num_input_events, nevents_read]: readers) {
        reader->closeFile();
    }
}
